# -*- coding:utf8 -*-
import os
from flask_wtf import Form
from wtforms.widgets import TextArea, HiddenInput
from wtforms import TextField, IntegerField, SelectMultipleField
from wtforms.validators import DataRequired, NumberRange, ValidationError


class EnvironmentForm(Form):
    name = TextField('name', validators=[DataRequired()])
    sort_order = IntegerField('sort_order', validators=[NumberRange(min=0, max=50)])
    execute_path = TextField('execute_path', validators=[DataRequired()])

    def validate_execute_path(form, field):
        if field.data and (not os.path.exists(field.data)):
            raise ValidationError('path {0} not found.'.format(field.data))


class JobForm(Form):
    name = TextField('name', validators=[DataRequired()])
    description = TextField('description', widget=TextArea(), validators=[DataRequired()])
    environment_id = IntegerField('environment_id', widget=HiddenInput(), validators=[DataRequired()])
    tasks = TextField('tasks',
        #widget=HiddenInput,
        validators=[DataRequired()])


class JobSetForm(Form):
    title = TextField('title', validators=[DataRequired()])
    description = TextField('description', widget=TextArea(), validators=[DataRequired()])
    job_id = SelectMultipleField(u'Job', coerce=int, validators=[DataRequired()])



# -*- coding:utf8 -*-
import os
import sys
import json
import urlparse

from oozappa import exec_fabric
from oozappa.records import Environment, Job, Jobset, JobsetJobList, ExecuteLog, get_db_session, init as init_db
from oozappa.forms import EnvironmentForm, JobForm, JobSetForm
from oozappa.fabrictools import FabricHelper

import logging
from uuid import uuid4
from datetime import datetime
import time

from flask import Flask, render_template, url_for, redirect, Response, abort, jsonify, request
from flask_sockets import Sockets

import filelock

logging.basicConfig(level=logging.WARN, format='%(asctime)s %(levelname)s %(message)s')
logger = logging.getLogger('oozappa')
# logger.setLevel(logging.DEBUG)

app = Flask('oozappa')
from oozappa.config import get_config
_settings = get_config()
try:
    app.config['SECRET_KEY'] = _settings.FLASK_SECRET_KEY
except AttributeError:
    should_create = raw_input('Create common environment here? [y/N]')
    if should_create == 'y':
        import shutil
        current = os.getcwd()
        shutil.copytree(os.path.join(os.path.dirname(__file__), '_structure', 'common'), 'common')
        with open('common/vars.py') as f:
            data = f.read()
        with open('common/vars.py', 'w') as f:
            f.write(data.format(uuid4().hex))
        print('create common directory. db file path and flask secret key are in common/vars.py.')
        sys.exit(0)
    print("couldn't find FLASK_SECRET_KEY in your ENVIRONMENT/vars.py")
    sys.exit(1)

init_db()
sockets = Sockets(app)


@app.template_filter('datetimefmt')
def _datetimefmt(d):
    return d and d.strftime("%Y/%m/%d %H:%M:%S") or ''

LOCK_FILE_NAME = 'oozappa_lock_file'

def _is_locked():
    return os.path.exists(LOCK_FILE_NAME)

def _locked_time():
    return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(os.stat(LOCK_FILE_NAME).st_ctime))

@sockets.route('/run_task')
def run_task(ws):
    message = ws.receive()
    if not message:
        return
    data = json.loads(message)
    if not data.get('fabric_path', None):
        ws.send('need fabric_path')
        return
    if not data.get('tasks', None):
        helper = FabricHelper(data.get('fabric_path'))
        ws.send('specify task(s) below.\n  ' + '\n  '.join(helper.task_list()))
        return
    lock = filelock.FileLock(LOCK_FILE_NAME)
    try:
        with lock.acquire(timeout=1):
            with exec_fabric(ws, data['fabric_path']) as executor:
                executor.doit(data['tasks'].split(' '))
    except filelock.Timeout as err:
        ws.send('''
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

            Lock aquired another jobset from {0}

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
'''.format(_locked_time()))
        return


@sockets.route('/run_jobset')
def run_jobset(ws):
    message = ws.receive()
    if not message:
        return
    lock = filelock.FileLock(LOCK_FILE_NAME)
    try:
        with lock.acquire(timeout=1):
            data = json.loads(message)
            jobset_id = data.get('jobset_id')
            session = get_db_session()
            jobset = session.query(Jobset).get(jobset_id)
            executelog = ExecuteLog()
            executelog.jobset = jobset
            session.add(executelog)
            session.commit()
            logfile = None
            if _settings.OOZAPPA_LOG:
                executelog.logfile = os.path.join(_settings.OOZAPPA_LOG_BASEDIR,
                  '{0}.log'.format(uuid4().hex))
                logfile = open(executelog.logfile, 'w')
            for job in jobset.jobs:
                with exec_fabric(ws, job.environment.execute_path) as executor:
                    if executor.doit(job.tasks.split(' '), logfile) != 0:
                        executelog.success = False
                        executelog.finished = datetime.now()
                        session.commit()
                        ws.send('\n&nbsp;\n&nbsp;\n')
                        ws.send('=' * 35)
                        ws.send('\n')
                        ws.send(u'[Oozappa:FAILES] Jobset: {0} in {1} seconds.'.format(jobset.title,
                          executelog.execute_time()).encode('utf8'))
                        ws.send('\n')
                        ws.send('=' * 35)
                        ws.send('\n&nbsp;\n&nbsp;\n')
                        break
            else:
                executelog.success = True
                executelog.finished = datetime.now()
                session.commit()
                ws.send('\n&nbsp;\n&nbsp;\n')
                ws.send('=' * 35)
                ws.send('\n')
                ws.send(u'[Oozappa:FINISHED] Jobset: {0} in {1} seconds.'.format(jobset.title,
                  executelog.execute_time()).encode('utf8'))
                ws.send('\n')
                ws.send('=' * 35)
                ws.send('\n&nbsp;\n&nbsp;\n')
            if logfile:
                logfile.close()
    except filelock.Timeout as err:
        ws.send('''
!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

            Lock aquired another jobset from {0}

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
'''.format(_locked_time()))


@app.route('/')
def index():
    session = get_db_session()
    jobset_list = session.query(Jobset).all()
    executelog_list = session.query(ExecuteLog).order_by(ExecuteLog.id.desc()).all()[:20]
    return render_template('index.html', jobset_list=jobset_list, executelog_list=executelog_list)


@app.route('/environments')
def environments():
    session = get_db_session()
    environment_list = session.query(Environment).all()
    return render_template('environment_list.html', environment_list=environment_list, form=EnvironmentForm())


@app.route('/environments/add', methods=['POST'])
def add_environment():
    form = EnvironmentForm()
    if form.validate_on_submit():
        session = get_db_session()
        #TODO
        environ = Environment()
        environ.name = form.name.data
        environ.sort_order = form.sort_order.data
        environ.execute_path = form.execute_path.data
        session.add(environ)
        session.commit()
        return redirect(url_for('environments'))
    return render_template('environment_list.html', form=form)


@app.route('/emvironments/<environment_id>/', methods=['GET', 'POST'])
def environment(environment_id):
    session = get_db_session()
    environ = session.query(Environment).get(environment_id)
    helper = FabricHelper(environ.execute_path)
    error_message = None
    form = JobForm()
    if not environ:
        abort(404)
    if form.validate_on_submit():
        not_found_tasks = helper.get_tasks(form.tasks.data.split(' ')).get('not_found')
        if len(not_found_tasks) == 0:
            job = Job()
            job.name = form.name.data
            job.description = form.description.data
            job.tasks = form.tasks.data
            job.environment = environ
            session.commit()
            return redirect(url_for('environment', environment_id=environment_id))
        error_message = 'tasks [{0}] not found.'.format(','.join(not_found_tasks))
    form.environment_id.data = environ.id
    return render_template('environment.html', form=form,
      environ=environ, environ_doc=helper.doc, error_message=error_message)

@app.route('/emvironments/<environment_id>/job/<job_id>/', methods=['GET', 'POST'])
def job(environment_id, job_id):
    session = get_db_session()
    environ = session.query(Environment).get(environment_id)
    job = session.query(Job).get(job_id)
    helper = FabricHelper(environ.execute_path)
    error_message = None
    form = JobForm(
        name=job.name, description=job.description,
        environment_id=job.environment_id,
        tasks=job.tasks
    )
    if not environ:
        abort(404)
    if form.validate_on_submit():
        not_found_tasks = helper.get_tasks(form.tasks.data.split(' ')).get('not_found')
        if len(not_found_tasks) == 0:
            job.name = form.name.data
            job.description = form.description.data
            job.tasks = form.tasks.data
            job.environment = environ
            session.commit()
            return redirect(url_for('environment', environment_id=environment_id))
        error_message = 'tasks [{0}] not found.'.format(','.join(not_found_tasks))
    form.environment_id.data = environ.id
    return render_template('job.html', form=form,
      environ=environ, job=job, error_message=error_message)

@app.route('/jobsets')
def jobsets():
    session = get_db_session()
    jobset_list = session.query(Jobset).all()
    return render_template('jobset_list.html', jobset_list=jobset_list)


@app.route('/jobsets/create', methods=['GET', 'POST'])
def create_jobset():
    form = JobSetForm()
    form.job_id.choices = Job.choices()
    session = get_db_session()
    if form.validate_on_submit():
        jobset = Jobset()
        jobset.title = form.title.data
        jobset.description = form.description.data
        session.add(jobset)
        for id in form.job_id.data:
            j = session.query(Job).get(id)
            jsl = JobsetJobList()
            jsl.jobset = jobset
            jsl.job = j
            session.add(jsl)
        session.commit()
        return redirect(url_for('jobsets'))
    return render_template('create_jobset.html', form=form, job_list=session.query(Job).all())

@app.route('/jobsets/reorder/', methods=['POST'])
def reorder_jobset():
    if request.headers['Content-Type'] != 'application/json':
        print(request.headers['Content-Type'])
        return jsonify(res='error'), 400
    jobset_id_ordered = [int(x) for x in urlparse.parse_qs(request.json).get('jobset[]', [])]
    session = get_db_session()
    for i, jobset_id in enumerate(jobset_id_ordered):
        _sort_order = i + 1
        _jobset = session.query(Jobset).get(jobset_id)
        if _jobset.sort_order != _sort_order:
            _jobset.sort_order = _sort_order
    session.commit()
    return jsonify(res='ok')

@app.route('/jobsets/<jobset_id>/', methods=['GET', 'POST'])
def jobset(jobset_id):
    session = get_db_session()
    jobset = session.query(Jobset).get(jobset_id)
    return render_template('jobset.html', jobset=jobset, job_list=session.query(Job).all(), executelog_list=jobset.execute_logs[:10])


@app.route('/get_execute_log/<id>/', methods=['GET'])
def get_execute_log(id):
    session = get_db_session()
    executelog = session.query(ExecuteLog).get(id)
    if not executelog.logfile:
        abort(404)
    with open(executelog.logfile, 'r') as f:
        logs = f.read()
    return Response(logs, mimetype='text/plain')

@app.route('/is_running_something.json', methods=['GET'])
def is_running_something():
    if _is_locked():
        return jsonify(dict(running=True, locked_time=_locked_time()))
    return jsonify(dict(running=False))

#gunicorn -t 3600 -k flask_sockets.worker oozappa.webui:app



# -*- coding:utf8 -*-
import os
import sys
from fabric.api import env
from fabric.main import load_fabfile
from fabric.contrib.files import upload_template as fabric_upload_template

import logging
logger = logging.getLogger('oozappa')


class FabricTask(object):
    def __init__(self, task_dict, m=None):
        self._task_dict = task_dict
        self._m = m

    @property
    def name(self):
        if self._m:
            return '{0}.{1}'.format(self._m, self._task_dict.name)
        return self._task_dict.name

    @property
    def description(self):
        return self._task_dict.__doc__ or u''

    def __str__(self):
        return self.name

    def __repr__(self):
        return self.__str__()


class FabricHelper(object):
    def __init__(self, path):
        _path = path
        if not path.endswith('fabric'):
            _path = os.path.join(path, 'fabfile')
        if 'fabfile' in sys.modules.keys():
            del sys.modules['fabfile']
        self.doc, _dict = load_fabfile(_path)[:2]
        self.task_dict = dict((x.name, FabricTask(x)) for x in _dict.values() if hasattr(x, 'name'))
        _modules = [k for k, m in _dict.items() if not hasattr(m, 'name')]
        for _m in _modules:
            self.task_dict.update(dict(('{0}.{1}'.format(_m, x.name), FabricTask(x, m=_m)) for x in _dict[_m].values() if hasattr(x, 'name')))
        self.directory = os.path.split(path)[0]

    def task_list(self):
        return self.task_dict

    def task(self, name):
        return self.task_dict[name]

    def get_tasks(self, task_list):
        tasks = self.task_list()
        result = dict(found={}, not_found=[])
        for t in task_list:
            if t in tasks:
                result.get('found')[tasks.get(t).name] = tasks.get(t)
            else:
                result.get('not_found').append(t)
        return result


def print_task_list(path):
    tasks = load_fabfile(path)
    task_dict = tasks[1]
    for task in task_dict.values():
        print(u'{0:10}: {1}'.format(task.name, task.__doc__))


def get_tasks(path, task_list):
    tasks = FabricHelper(path).task_list()
    result = dict(found={}, not_found=[])
    for t in task_list:
        if t in tasks:
            result.get('found')[tasks.get(t).name] = tasks.get(t)
        else:
            result.get('not_found').append(t)
    return result


TEMPLATES_DIRNAME = 'templates'


def upload_template(filename, destination, context=None,
    template_dir=None, use_sudo=False, backup=False, mirror_local_mode=False,
        mode=None):
    '''Search filename in ENVIRONMENT_DIR/templates at first and then common/templates if template_dir is not supplied.
      Call fabric.contrib.file.upload_template passing directory that is found filename, or None.
      if template_dir passed, this function just call fabric.contrib.file.upload_template as usual.

      *) backup default option is False

      *) use_jinja option is always True and you can't pass use_jinja option to this function.
    '''
    _template_dir = template_dir
    if not _template_dir:
        call_path = os.getcwd()
        common_path = os.path.abspath(os.path.join(call_path, '..', 'common'))
        if os.path.exists(os.path.join(call_path, TEMPLATES_DIRNAME, filename)):
            logger.debug('found called path')
            _template_dir = os.path.join(call_path, TEMPLATES_DIRNAME)
        elif os.path.exists(os.path.join(common_path, TEMPLATES_DIRNAME, filename)):
            logger.debug('found common path')
            _template_dir = os.path.join(common_path, TEMPLATES_DIRNAME)
        else:
            logger.warn('NOT FOUND {0}'.format(os.path.join(call_path, TEMPLATES_DIRNAME, filename)))
            logger.warn('NOT FOUND {0}'.format(os.path.join(common_path, TEMPLATES_DIRNAME, filename)))
    return fabric_upload_template(filename, destination, context=context, use_jinja=True,
        template_dir=_template_dir, use_sudo=use_sudo, backup=backup, mirror_local_mode=mirror_local_mode,
        mode=mode)


def only_once(f):
    '''Task decorator. Run task only once while executing multiple task'''
    def _f(*args, **kwargs):
        flag_format = 'only_once_function_{0}'
        try:
            if not env.get(flag_format.format(f.__name__)):
                return f(*args, **kwargs)
            logger.info('[ONLY ONCE] pass function [{0}]'.format(f.__name__))
        finally:
            env[flag_format.format(f.__name__)] = True
    return _f



# -*- coding:utf8 -*-
import sys
import os
from copy import deepcopy
import collections

import logging
logger = logging.getLogger('oozappa')


def _update(org, opt):
    for k, v in opt.iteritems():
        if isinstance(v, collections.Mapping):
            r = _update(org.get(k, OozappaSetting()), v)
            org[k] = r
        else:
            org[k] = opt[k]
    return org


class OozappaSetting(dict):
    '''dict like object. accessible with dot syntax.
  >>> settings = OozappaSetting(
  ...   spam = '123',
  ...   egg = 123
  ... )
  >>> assert(settings.spam == '123')
  >>> assert(settings.egg == 123)
  >>> settings.ham = 123.0
  >>> assert(settings.ham == 123.0)
  >>> s2 = OozappaSetting(dict(spam=456))
  >>> settings.update(s2)
  >>> assert(settings.spam == 456)
  >>> assert(settings.ham == 123.0)
    '''
    
    def __init__(self, *args, **kwargs):
        for d in args:
            if isinstance(d, collections.Mapping):
                self.update(d)
        for key, value in kwargs.items():
            self[key] = value

    def __setattr__(self, key, value):
        if isinstance(value, collections.Mapping):
            self[key] = OozappaSetting(value)
        else:
            self[key] = value

    def __getattr__(self, key):
        try:
            return self[key]
        except:
            object.__getattribute__(self, key)

    def update(self, opt):
        self = _update(self, opt)


def common_base_path():
    if os.path.exists(
        os.path.abspath(
            os.path.join(os.getcwd(), 'fabfile'))) or os.path.exists(
        os.path.abspath(
            os.path.join(os.getcwd(), 'fabfile.py'))):
        return os.path.abspath(os.path.join(os.getcwd(), '..'))
    else:
        return os.getcwd()


def get_config():
    path_added = False
    _common_base_path = common_base_path()
    logger.debug('COMMON_BASE_PATH: {0}'.format(_common_base_path))
    try:
        if not _common_base_path in sys.path:
            sys.path.insert(0, _common_base_path)
            path_added = True
        from common.vars import settings as common_settings
        _settings = deepcopy(common_settings)
        if path_added:
            sys.path.remove(_common_base_path)
    except ImportError:
        _settings = OozappaSetting()
    try:
        from vars import settings
        _settings.update(settings)
    except ImportError:
        pass
    return _settings


def procure_common_functions():
    u'''insert common/functions to sys.path'''
    _common_function_path = os.path.join(common_base_path(), 'common', 'functions')
    if not _common_function_path in sys.path:
        if not os.path.exists(_common_function_path):
            logging.warn('common_function directory not exists? {0}'.format(_common_function_path))
        sys.path.insert(0, _common_function_path)
        logger.info('common_function directory added to sys.path')



# -*- coding:utf8 -*-
import os
import sys
import subprocess
import time
import logging

logging.basicConfig(level=logging.INFO, format='%(asctime)s %(levelname)s %(message)s')
logger = logging.getLogger('oozappa')

version_info = (0, 8, 5)
__version__ = ".".join([str(v) for v in version_info])


class exec_fabric:
    def __init__(self, wsckt, path):
        self.wsckt = wsckt
        self.initial_path = os.getcwd()  # TODO
        if not os.path.isabs(path):
            self.exec_path = os.path.join(self.initial_path, path)
        else:
            self.exec_path = path
        self.path_appended = False

    def __enter__(self):
        os.chdir(self.exec_path)
        if not self.exec_path in sys.path:
            sys.path.insert(0, self.exec_path)
            self.path_appended = True
        return self

    def __exit__(self, type, value, traceback):
        os.chdir(self.initial_path)
        if self.path_appended:
            sys.path.remove(self.exec_path)

    def doit(self, fabric_commands=[], logfile=None):
        start_time = time.time()
        execute_commands = '<<doit from running websocket: fab {0}>>\n'.format(' '.join(fabric_commands))
        logger.debug(execute_commands)
        if logfile:
            logfile.write(execute_commands)
        p = subprocess.Popen(["fab"] + fabric_commands,
          stdout=subprocess.PIPE, stderr=subprocess.STDOUT, universal_newlines=True)
        stderr = p.stdout
        while True:
            line = stderr.readline()
            if not line:
                break
            self.wsckt.send(line)
            if logfile:
                logfile.write(line)
            # print(line.strip())
        self.wsckt.send('takes {0:.2f} sec'.format(time.time() - start_time))
        return p.wait()



# -*- coding:utf8 -*-
from oozappa.config import OozappaSetting
_ = OozappaSetting

import tempfile

settings = _(
        # OOZAPPA_DB: specify sqlite database stored file path.
        OOZAPPA_DB = '{OOZAPPA_DB}',
        # OOZAPPA_LOG: set True if you need output log to file.
        OOZAPPA_LOG = True,
        # OOZAPPA_BASEDIR: specify output log file directory.
        OOZAPPA_LOG_BASEDIR = '{OOZAPPA_LOG_BASEDIR}',
        # FLASK_SECRET_KEY: flask thing.
        FLASK_SECRET_KEY = '{FLASK_SECRET_KEY}',
)



# -*- coding:utf8 -*-
import os
import sys
import logging
import shutil
from uuid import uuid4

logger = logging.getLogger('oozappa')

from oozappa.config import get_config
_settings = get_config()


from oozappa import __version__

DEFAULT_OOZAPPA_PATH = '/tmp/oozappa'
DEFAULT_OOZAPPA_DATAPATH = '{DEFAULT_OOZAPPA_PATH}/data.sqlite'.format(
    DEFAULT_OOZAPPA_PATH=DEFAULT_OOZAPPA_PATH)


def _ensure_directory(pth):
    if not os.path.exists(pth):
        if raw_input('Create directory or exit? "{0}" [y/N] : '.format(pth)).lower() == 'y':
            os.makedirs(pth)
        else:
            print('aborted')
            sys.exit(-1)
    if not os.path.isdir(pth):
        print('{0} is not directory.'.format(pth))
        print('aborted')
        sys.exit(-1)
    if not os.access(pth, os.W_OK | os.X_OK):
        print('directory "{0}" is not writable. check permission.'.format(pth))
        sys.exit(1)


def _create_database_path(sqlite_stored_path):
    if sqlite_stored_path == '':
        sqlite_stored_path = DEFAULT_OOZAPPA_DATAPATH
    else:
        if os.path.dirname(sqlite_stored_path) == '':
            sqlite_stored_path = os.path.join(os.path.abspath(os.path.curdir), sqlite_stored_path)
        sqlite_stored_path = os.path.abspath(sqlite_stored_path)
        _ensure_directory(sqlite_stored_path)
        sqlite_stored_path = 'sqlite:///{0}'.format(sqlite_stored_path)
    return sqlite_stored_path


def _create_logfile_stored_path(log_file_path):
    if log_file_path == '':
        log_file_path = DEFAULT_OOZAPPA_PATH
    _ensure_directory(log_file_path)
    return log_file_path


def init(args):
    try:
        _settings.FLASK_SECRET_KEY
    except AttributeError:
        should_create = raw_input('Create common environment here? [y/N] : ').lower()
        if should_create == 'y':
            sqlite_stored_path = _create_database_path(raw_input(
                'Sqlite database stored path. [{DEFAULT_OOZAPPA_DATAPATH}] : '.format(
                    DEFAULT_OOZAPPA_DATAPATH=DEFAULT_OOZAPPA_DATAPATH))
            )
            log_file_path = _create_logfile_stored_path(
                raw_input('Log files stored path. [{DEFAULT_OOZAPPA_PATH}] : '.format(
                    DEFAULT_OOZAPPA_PATH=DEFAULT_OOZAPPA_PATH))
            )
            shutil.copytree(os.path.join(os.path.dirname(__file__), '_structure', 'common'), 'common')
            with open('common/vars.py') as f:
                data = f.read()
            with open('common/vars.py', 'w') as f:
                f.write(data.format(FLASK_SECRET_KEY=uuid4().hex,
                  OOZAPPA_DB=sqlite_stored_path,
                  OOZAPPA_LOG_BASEDIR=log_file_path))
            print('created common directory. db/log file path and flask secret key are in common/vars.py.')
            sys.exit(0)
        print("couldn't find FLASK_SECRET_KEY in your ENVIRONMENT/vars.py")
        sys.exit(1)
    else:
        print('Oozappa structure exits already.')


def create_environment(args):
    environ_names = args.names
    if len(environ_names) == 0:
        while 1:
            x = raw_input('environ_name : ').strip()
            if len(x) > 1 and 'common' != x.lower():
                environ_names.append(x)
                break
    for environ_name in environ_names:
        if os.path.exists(environ_name):
            logger.warn('{0} is exists and do nothing.'.format(environ_name))
            continue
        shutil.copytree(os.path.join(os.path.dirname(__file__), '_structure', '_environment'), environ_name)
        logger.info('create environment : {0}'.format(environ_name))


def print_version(args):
    print('Oozappa/{0}'.format(__version__))

import argparse


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(help='commands')
    init_parser = subparsers.add_parser('init', help='Initialize. Create oozappa base structure.')
    init_parser.set_defaults(func=init)
    create_environemnt_parser = subparsers.add_parser('create_environment', help='Create oozappa environment structure.')
    create_environemnt_parser.add_argument('--names', nargs='+', default=[])
    create_environemnt_parser.set_defaults(func=create_environment)
    version_parser = subparsers.add_parser('version', help='show Oozappa version')
    version_parser.set_defaults(func=print_version)
    args = parser.parse_args()
    args.func(args)



#!/usr/bin/env python
# -*- coding:utf8 -*-
u'''
Oozappa
============================================================

Fabric task runner and helper. Execute tasks via web browser.

oozappa is 大雑把.

Change logs.
-------------------------------------------

0.8.5
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Adding new feature, modify Jobsets sort order on jobsets page.

Adding new feature, modify a Job data on environ page.

Improve Fabric task's behavior better when failed.

You need db migration if you use 0.8.4 or below.

Download zip archive from https://github.com/tsuyukimakoto/oozappa/raw/master/migration/to_0.8.5.zip

::

  $ unzip to_0.8.5.zip && cd oozappa_to_0_8_5

  If you have not alembic installed. You need install it using pip or any way you like.
  Open alembic.ini and modify sqlalchemy.url to indicate your own oozappa's db. then

  $ alembic upgrade 1e26220da128

0.8.4
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Jobset runs exclusively from webui.

0.8.3
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

A zappa command added and removed oozappa.create_environment module.

0.8.2
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Oozappa finds tasks from imported modules, not recursive.

Install
-------------------------------------------

Install from Cheese Shop (pypi).

::

    $ pip install oozappa

If you use Xcode 5.1(above) and failed with **clang: error: unknown argument: '-mno-fused-madd'** , export flags before install.
::

    $ export CPPFLAGS=-Qunused-arguments
    $ export CFLAGS=-Qunused-arguments

Oozappa data models.
-------------------------------------------

Environment
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Environment is a category that has fibfile directory and vars.py .

A problem about fabric with large project is, too many fabric tasks and complicated task orders.

So you should separate fabfile into domain category. That is Environment.

Job
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Job is like a normal fabric execution command.

note::

 fab task1 task2 is a job.

Jobset
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Jobset is oozappa operation unit.

Jobset can contain multiple job, even extend over environments.

Oozappa fabric structures.
-------------------------------------------

See sample oozappa project( https://github.com/tsuyukimakoto/oozappa/tree/master/sample/ops ).
::

    .
    ├── common
    │   ├── __init__.py
    │   ├── files
    │   ├── functions
    │   │   ├── common_multiple_fabric_environment.py
    │   ├── templates
    │   │   └── sample_a.txt
    │   └── vars.py
    ├── construction
    │   ├── fabfile
    │   │   ├── __init__.py
    │   │   └── cloud.py
    │   ├── templates
    │   └── vars.py
    ├── deployment
    │   ├── fabfile
    │   │   ├── __init__.py
    │   ├── templates
    │   └── vars.py
    ├── production
    │   ├── fabfile
    │   │   ├── __init__.py
    │   ├── templates
    │   └── vars.py
    └── staging
        ├── fabfile
        │   ├── __init__.py
        ├── templates
        │   └── sample_a.txt
        └── vars.py

**common** is reserved directory. **construction** and others are environment directory. These names except common are just example.

vars
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

common and each environment's vars.py might have oozappa.config.OozappaSetting instance named settings.
OozappaSetting is dict like object.

**common.vars.setting** is updated by executed environment's **vars.setting** , so you can set base configuration to common.vars.setting and environment's one overwrite it.

Check printsetting task on staging environment.

You can run fabric task within environment directory as usual.
::

    $ cd sample/ops/staging
    $ fab printsetting
    {'instance_type': 't1.micro', 'domain': 'localhost', 'sample_template_vars': {'sample_a': {'key_a_2': "a's 2 value from common.vars", 'key_a_1': "a's 1 value from stging.vars"}}, 'email': 'mtsuyuki at gmail.com'}

See common/vars.py and staging/vars.py .

templates
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Same as vars, **oozappa.fabrictools.upload_template** search template. upload_template is almost same as **fabric.contrib.files.upload_template** . oozappa's upload_template doesn't accept use_jinja, because oozappa's upload_template pass use_jinja=True to fabric.contrib.files.upload_template.

Jinja2 has inheritance template system and search template from multipul paths. fabric's upload_template accept only one template_dir string not list. **fabric doesn't assume multiplu environment** , so it's reasonable.
Because of this, oozappa's upload_template search template path is limited only one template_dir that found filename.
It mean that you can't store child template and parent template separately.

common/functions
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Call **oozappa.config.procure_common_functions** () and add commons/functions directory to sys.path for convinient to using on multiple fabric environment.

Run fabric task via web browser.
-----------------------------------------------------------

Change directory to outside environment directory.
::

    $ cd ..
    $ ls
    common    production  staging
    $ gunicorn -t 3600 -k flask_sockets.worker oozappa.webui:app

Running oozappa:app creates **/tmp/oozappa.sqlite** .

Open your web browser and browse http://localhost:8000/ .

Rapid execution
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Modify *Run fabric in raw*'s left hand side input to **staging** and click *run_tasks*. You can see what tasks exists.

Then input **ls ps** to right hand side input and click *run_tasks* .

That's it.

Better way using sample
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

register environment to db.
___________________________

* Click environment button via top menu.

* Add new Environment

  * name: constructiton

  * sort_order: 1

  * execute_path: constructiton

* Add 3 more.

.. image:: https://dl.dropboxusercontent.com/u/382460/oozappa/readme/environments.png
  :alt: environments

create job in each environments.
_________________________________

* Click environment you created

* Create new Job.

  * Click task from Possible tasks in order

.. image:: https://dl.dropboxusercontent.com/u/382460/oozappa/readme/create_job.png
  :alt: job

create jobset
___________________________

* Click jobset button via top menu.

* Click jobs you'd like to execute once.

.. image:: https://dl.dropboxusercontent.com/u/382460/oozappa/readme/create_jobset.png
  :alt: jobset


run jobset
___________________________

* Click navigation button or jobset button via top menu.

* Click jobset you'd like to execute.

* Click *run jobset* button.
* Running log displays **Running log**.

* Reload page when jobset done. or Go to top(via navigation button)

  * You see Execute Logs and show raw log when you click success (or fail).

.. image:: https://dl.dropboxusercontent.com/u/382460/oozappa/readme/jobset.png
  :alt: running jobset

How to create your own
-------------------------------------------

Create common directory
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Change directory your own oozappa.
::

    $ mkdir devops
    $ cd devops

Then run zappa command.
::

    $ zappa init
    Create common environment here? [y/N] : y
    Sqlite database stored path. [/tmp/oozappa/data.sqlite] :
    Log files stored path. [/tmp/oozappa] :
    Create directory or exit? "/tmp/oozappa" [y/N] : y
    created common directory. db/log file path and flask secret key are in common/vars.py.

.. attention:: Default stored path is not for production use.
   You should input your own file/directory path. Or data/results disapear when you reboot your machine or server.

How to change settings
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Open `common/vars.py` and change settings.

* **OOZAPPA_DB**

  sqlite's data store path.

* **OOZAPPA_LOG_BASEDIR**

  Jobset execute log store directory path.

Create environment
^^^^^^^^^^^^^^^^^^^^^^^^^^^^

Run `zappa` command with names option.
::

  $ zappa create_environment --names construction deployment
  2014-04-20 16:43:26,543 INFO create environment : construction
  2014-04-20 16:43:26,544 INFO create environment : deployment

Then you can write fabfile normally and execute via oozappa.
'''
import setuptools
from distutils.core import setup

from oozappa import __version__

setup(name='oozappa',
      version=__version__,
      description='Fabric task runner and helper. Executes and manages tasks via web browser.',
      author='makoto tsuyuki',
      author_email='mtsuyuki@gmail.com',
      url='https://github.com/tsuyukimakoto/oozappa',
      zip_safe=False,
      entry_points='''
            [console_scripts]
            zappa = oozappa.zappa:main
            ''',
      long_description=__doc__,
      classifiers=['Development Status :: 3 - Alpha',
          'Environment :: Console',
          'Environment :: Web Environment',
          'Intended Audience :: Developers',
          'Intended Audience :: System Administrators',
          'License :: OSI Approved :: MIT License',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX',
          'Operating System :: Unix',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2.7',
          'Topic :: Software Development',
          'Topic :: Software Development :: Build Tools',
          'Topic :: Software Development :: Libraries',
          'Topic :: Software Development :: Libraries :: Python Modules',
          'Topic :: System :: Clustering',
          'Topic :: System :: Software Distribution',
          'Topic :: System :: Systems Administration',
      ],
      packages=['oozappa', 'oozappa.records',],
      package_data={'oozappa': ['_structure/_environment/vars.py','_structure/_environment/*/*', '_structure/_environment/templates/*',
        '_structure/common/vars.py','_structure/common/__init__.py','_structure/common/*/*', 'static/css/*', 'templates/*']},
      include_package_data=True,
      install_requires=['Fabric>=1.8.3','Flask-WTF>=0.9.5','Flask-SQLAlchemy>=1.0','Flask-Sockets>=0.1',
        'Jinja2>=2.7.2', 'Pygments>=1.6', 'gevent-websocket>=0.9.3','gunicorn>=18.0','filelock>=0.2.0'],
     )



# -*- coding:utf8 -*-
from fabric.api import task, local, run, sudo, env

from oozappa.config import get_config, procure_common_functions
_settings = get_config()
procure_common_functions()

# your own task below



# -*- coding:utf8 -*-
from fabric.main import load_fabfile
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Table, Column, Integer, String, DateTime, Boolean, ForeignKey
from sqlalchemy.orm import relationship, backref

import os
import sys
from datetime import datetime

Base = declarative_base()

from functools import wraps

import logging
logger = logging.getLogger('oozappa')

from oozappa.fabrictools import FabricHelper


def need_task(f):
    @wraps(f)
    def inner(self, *args, **kwargs):
        if not hasattr(self, 'task_dict'):
            if 'fabfile' in sys.modules.keys():
                del sys.modules['fabfile']
            self.task_dict = FabricHelper(os.path.join(self.execute_path, 'fabfile')).task_list()
        return f(self, *args, **kwargs)
    return inner


class Environment(Base):
    '''Target, normally this indicates fabfile.py or fabfile directory'''
    __tablename__ = 'environment'

    id = Column(Integer, primary_key=True)
    name = Column(String, unique=True)
    sort_order = Column(Integer)
    execute_path = Column(String)

    def __init__(self):
        self.task_dict = None

    @need_task
    def task_list(self):
        return self.task_dict.values()

    def __repr__(self):
        return "<Environment(name='%s', sort_order='%d', execute_path='%s')>" % (
            self.name, self.sort_order, self.execute_path)


class Jobset(Base):
    __tablename__ = 'jobset'
    id = Column(Integer, primary_key=True)
    title = Column(String)
    description = Column(String)
    sort_order = Column(Integer, server_default="1")

    __mapper_args__ = {'order_by': 'Jobset.sort_order'}

    @property
    def jobs(self):
        return [j.job for j in self.jobsetlist]


class Job(Base):
    __tablename__ = 'job'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    description = Column(String)
    environment_id = Column(Integer, ForeignKey('environment.id'))
    tasks = Column(String)
    environment = relationship("Environment", backref=backref('jobs', order_by=id))

    @staticmethod
    def choices():
        return [(x.id, x.name) for x in get_db_session().query(Job).all()]


class JobsetJobList(Base):
    __tablename__ = 'jobset_job'
    id = Column(Integer, primary_key=True)
    jobset_id = Column(Integer, ForeignKey('jobset.id'))
    job_id = Column(Integer, ForeignKey('job.id'))

    jobset = relationship("Jobset", backref=backref('jobsetlist', order_by=id))
    job = relationship("Job", backref=backref('jobsetlist', order_by=id))


class ExecuteLog(Base):
    __tablename__ = 'execute_log'
    id = Column(Integer, primary_key=True)
    success = Column(Boolean)
    started = Column(DateTime)
    finished = Column(DateTime)
    logfile = Column(String)
    jobset_id = Column(Integer, ForeignKey('jobset.id'))
    jobset = relationship("Jobset", backref=backref('execute_logs', order_by=id.desc()))

    def __init__(self):
        self.success = False
        self.started = datetime.now()

    def execute_time(self):
        if self.started and self.finished:
            return '{0:08.3f}'.format((self.finished - self.started).total_seconds())
        return ''

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from oozappa.config import get_config
_settings = get_config()


def get_db_session():
    engine = create_engine(_settings.OOZAPPA_DB, echo=False)
    Session = sessionmaker(bind=engine)
    return Session()


def init():
    from sqlalchemy import create_engine
    engine = create_engine(_settings.OOZAPPA_DB, echo=False)
    metadata = Base.metadata  # this knows table mapping class.
    metadata.create_all(engine)  # create table if not exists.
